#include <UrsAds1x15.h>

#if defined(ESP8266)
#include <UrsTwi.h>
UrsTwi twi;
#elif defined(__AVR__) 
#include <Wire.h>
#define ICACHE_RAM_ATTR
#define twi Wire
#endif 

const    uint8_t maxData = 33;   // Anzahl Messwerte
                                 // Ergibt 32 valide Werte. Wert 0 wird verworfen, ist hufig ein Aureier.
volatile int16_t data[maxData];  // Array zur Aufnahme der Daten
volatile uint8_t index = 0;      // Index fr data

#if defined(ESP8266)
UrsAds1015 ads(twi); // 12-Bit ADC
//UrsAds1115 ads(twi);   // 16-Bit ADC
// SCL: D1
// SDA: D2
const uint8_t alertPin = D4;  // Interrupt-Pin
#elif defined(__AVR__)
UrsAds1015 ads; // 12-Bit ADC
//UrsAds1115 ads;   // 16-Bit ADC
// SCL: A5
// SDA: A4
const uint8_t alertPin = 2;  // Interrupt-Pin (D2, INT0)
#endif 

// ISR
void ICACHE_RAM_ATTR ADCISR() {
   if (index < maxData) {
      data[index++] = ads.readContinuousModeRaw();
   }
}

// Startet den Messvorgang
// Wird zweimal benutzt, sorgt fr gleiche Parameter bei beiden Aufrufen.
void startContinuousMode(void) {
   index = 0;
   ads.startContinuousMode(Ads1x15Mux::Single_0, ADCISR, alertPin); // Messvorgang starten
}

void setup(void) {
   ads.reset();
   Serial.begin(115200);
   Serial.println(F("\n\n----------------------------\nADS1015\n"));
   delay(1000);

   twi.begin(); // IC-Interface starten

   // Im Continous-Mode und bei 3300 SPS beim ADS1015 muss die Takfrequenz auf min. 400kHz gesetzt weden.
   // Bei der Standardeinstellung 100kHz dauert das Einlesen gut 320us + die zustzlich auszufhrenden Anweisungen.
   // Es stehen aber nur max. 300us zur Verfgung. Der nchste Interrupt erfolgt dann bevor die ISR des
   // des vorhergehenden beendet wurde.
   twi.setClock(400000L); // IC-Taktfrequenz: 400 kHz
                          // setClock() muss N A C H begin() aufgerufen werden.
                          // begin() setzt die Transferate auf 100kHz.


   ads.setSps(Ads1015Sps::Sps2400);   // 3300 Messungen pro Sekude
  // ads.setMaxSps(); // 3300 Messungen pro Sekude beim ADS1015, 860 SPS beim ADS1115

   ads.setGain(Ads1x15Gain::Gain_16); // Hchste Empfindlichkeit

   Serial.print(F("Sample rate: ")); Serial.print(ads.getSamplesPerSeconds()); Serial.println(" SPS");
   Serial.print(F("Gain:        ")); Serial.println(ads.getGainString());
   Serial.print(F("mV per Bit:  ")); Serial.println(ads.getMilliVoltsPerBit());
   startContinuousMode(); // Messvorgang starten
}

void loop(void) {
   if (index < maxData)
      return; // Messung luft noch

   ads.stopContinuousMode(); // Messung stoppen
   Serial.println(F("\n\n----------------------------\n"));

   float sum = 0;
   for (int i = 1; i < maxData; i++) { // Messwerte ausgeben, ohne 1. Wert, der ist hufig ein Ausreier
      Serial.print(i); Serial.print("\t");
      float mv = ads.getMilliVoltsPerBit()*data[i];
      Serial.print(mv); Serial.println("\tmV");
      sum += mv;
   }

   Serial.println(F("-----------------"));
   Serial.print(F("Mean:\t"));  Serial.print(sum / (maxData - 1)); Serial.println("\tmV");
   delay(5000); // 5 Sekunden Pause

   // Messung neu starten
   startContinuousMode();
}
