// UrsAds1x15
// Erweiterte Alternative zu der Adafruit-Library: // https://www.adafruit.com/product/1083
//
// Autor: http://UllisRoboterSeite.de
// Doku:  http://bienonline.magix.net/public/esp8266-UrsWiFi.html

#include <UrsAds1x15.h>

#if defined(__AVR__) 
#define _twi Wire
#define ICACHE_RAM_ATTR
#endif 
/*=========================================================================
POINTER REGISTER
-----------------------------------------------------------------------*/
#define ADS1X15_REG_POINTER_MASK        (0x03)
#define ADS1X15_REG_POINTER_CONVERT     (0x00)
#define ADS1X15_REG_POINTER_CONFIG      (0x01)
#define ADS1X15_REG_POINTER_LOWTHRESH   (0x02)
#define ADS1X15_REG_POINTER_HITHRESH    (0x03)
/*=========================================================================*/

/*=========================================================================
CONFIG REGISTER
-----------------------------------------------------------------------*/
#define ADS1X15_REG_CONFIG_OS_MASK      (0x8000)
#define ADS1X15_REG_CONFIG_OS_SINGLE    (0x8000)  // Write: Set to start a single-conversion
#define ADS1X15_REG_CONFIG_OS_BUSY      (0x0000)  // Read: Bit = 0 when conversion is in progress
#define ADS1X15_REG_CONFIG_OS_NOTBUSY   (0x8000)  // Read: Bit = 1 when device is not performing a conversion

#define ADS1X15_REG_CONFIG_MUX_MASK     (0x7000)  // Maske fr Multiplexer
#define ADS1X15_REG_CONFIG_PGA_MASK     (0x0E00)  // Maske fr Gain

#define ADS1X15_REG_CONFIG_MODE_MASK    (0x0100)
#define ADS1X15_REG_CONFIG_MODE_CONTIN  (0x0000)  // Continuous conversion mode
#define ADS1X15_REG_CONFIG_MODE_SINGLE  (0x0100)  // Power-down single-shot mode (default)

#define ADS1X15_REG_CONFIG_DR_MASK      (0x00E0)  // Maske fr SamplesPerSecond

#define ADS1X15_REG_CONFIG_CMODE_MASK   (0x0010)
#define ADS1X15_REG_CONFIG_CMODE_TRAD   (0x0000)  // Traditional comparator with hysteresis (default)
#define ADS1X15_REG_CONFIG_CMODE_WINDOW (0x0010)  // Window comparator

#define ADS1X15_REG_CONFIG_CPOL_MASK    (0x0008)
#define ADS1X15_REG_CONFIG_CPOL_ACTVLOW (0x0000)  // ALERT/RDY pin is low when active (default)
#define ADS1X15_REG_CONFIG_CPOL_ACTVHI  (0x0008)  // ALERT/RDY pin is high when active

#define ADS1X15_REG_CONFIG_CLAT_MASK    (0x0004)  // Determines if ALERT/RDY pin latches once asserted
#define ADS1X15_REG_CONFIG_CLAT_NONLAT  (0x0000)  // Non-latching comparator (default)
#define ADS1X15_REG_CONFIG_CLAT_LATCH   (0x0004)  // Latching comparator

#define ADS1X15_REG_CONFIG_CQUE_MASK    (0x0003)
#define ADS1X15_REG_CONFIG_CQUE_1CONV   (0x0000)  // Assert ALERT/RDY after one conversions
#define ADS1X15_REG_CONFIG_CQUE_2CONV   (0x0001)  // Assert ALERT/RDY after two conversions
#define ADS1X15_REG_CONFIG_CQUE_4CONV   (0x0002)  // Assert ALERT/RDY after four conversions
#define ADS1X15_REG_CONFIG_CQUE_NONE    (0x0003)  // Disable the comparator and put ALERT/RDY in high state (default)
/*=========================================================================*/

void UrsAds1x15::writeRegister(uint8_t reg, uint16_t value) {
  _twi.beginTransmission(_i2cAddress);
  _twi.write(reg);
  _twi.write((uint8_t)(value >> 8));
  _twi.write((uint8_t)(value & 0xFF));
  _twi.endTransmission(true);
}

uint16_t ICACHE_RAM_ATTR UrsAds1x15::readRegister(uint8_t reg) {
  _twi.beginTransmission(_i2cAddress);
  _twi.write(reg);
  _twi.endTransmission(true);
  _twi.requestFrom(_i2cAddress, (uint8_t) 2, (uint8_t)true);
  return ((_twi.read() << 8) | _twi.read());
}

// Constructor UrsAds1x15
#if defined(ESP8266)
UrsAds1x15::UrsAds1x15(UrsTwi& twi, Ads1x15Address i2cAddress) : _twi(twi) {
#elif defined(__AVR__) 
UrsAds1x15::UrsAds1x15(Ads1x15Address i2cAddress) {
#endif 
  _i2cAddress = (uint8_t) i2cAddress;
  setGain(Ads1x15Gain::Gain_2_3); /* +/- 6.144V range (limited to VDD +0.3V max!) */
}

Ads1x15Gain UrsAds1x15::getGain() const {
  return _gain;
}

void UrsAds1x15::reset() {
   static const uint8_t general_i2c_addr = 0x00;
   static const uint8_t ads1115_i2c_reset = 0x06;
   _twi.beginTransmission(general_i2c_addr);
   _twi.write(ads1115_i2c_reset);
   _twi.endTransmission(true);
}

void UrsAds1x15::setGain(const Ads1x15Gain gain) {
  _gain = gain;
  switch (gain) {
  case Ads1x15Gain::Gain_2_3: _mVoltsPerBit = 0.1875;    break;
  case Ads1x15Gain::Gain_1:   _mVoltsPerBit = 0.125;     break;
  case Ads1x15Gain::Gain_2:   _mVoltsPerBit = 0.0625;    break;
  case Ads1x15Gain::Gain_4:   _mVoltsPerBit = 0.03125;   break;
  case Ads1x15Gain::Gain_8:   _mVoltsPerBit = 0.015625;  break;
  case Ads1x15Gain::Gain_16:  _mVoltsPerBit = 0.0078125; break;
  }
}

int16_t UrsAds1x15::readAdcRaw(Ads1x15Mux channel) {
  if (_isContinuousMode)
    return 0;

  // Config Register zusammenstellen
  uint16_t config = ADS1X15_REG_CONFIG_CQUE_NONE |    // Disable the comparator (default val)
                    ADS1X15_REG_CONFIG_CLAT_NONLAT |  // Non-latching (default val)
                    ADS1X15_REG_CONFIG_CPOL_ACTVLOW | // Alert/Rdy active low   (default val)
                    ADS1X15_REG_CONFIG_CMODE_TRAD |   // Traditional comparator (default val)
                    ADS1X15_REG_CONFIG_MODE_SINGLE |  // Single-shot mode (default)
                    ADS1X15_REG_CONFIG_OS_SINGLE;     // Set 'start single-conversion' bit

  config |= (uint16_t)_SpsCode; // Messrate/-dauer auswhlen
  config |= (uint16_t)_gain;    // Vorverstrker auswhlen
  config |= (uint16_t)channel;  // Kanal auswhlen

  // Config Register ausgeben
  writeRegister(ADS1X15_REG_POINTER_CONFIG, config);

  // Auf das Ende der Messung warten
  delay(_conversionDelay);

  // Messwert auslesen
  return (int16_t)readRegister(ADS1X15_REG_POINTER_CONVERT);
}

void UrsAds1x15::startContinuousMode(const Ads1x15Mux mux, void(*ISR)(void), const uint8_t alertPin, const bool usePullUpForAlertPin) {
  uint16_t config = ADS1X15_REG_CONFIG_CQUE_1CONV |   // Disable the comparator
                    ADS1X15_REG_CONFIG_CLAT_NONLAT |  // Non-latching 
                    ADS1X15_REG_CONFIG_CPOL_ACTVLOW | // Alert/Rdy active low
                    ADS1X15_REG_CONFIG_CMODE_TRAD |   // Traditional comparator
                    ADS1X15_REG_CONFIG_MODE_CONTIN;   // Continuous conversion mode

  config |= (uint16_t)_SpsCode; // Messrate/-dauer auswhlen
  config |= (uint16_t)_gain;    // Vorverstrker auswhlen
  config |= (uint16_t)mux;      // Kanal auswhlen

  writeRegister( ADS1X15_REG_POINTER_HITHRESH, 0x8000); // High- & Low-Threshold-Register fr RDY konfigurieren
  writeRegister( ADS1X15_REG_POINTER_LOWTHRESH, 0);

  // Interrupt initialisieren
  if (usePullUpForAlertPin)
    pinMode(alertPin, INPUT_PULLUP); // PullUp auf dem Interrupt-Pin
  else
    pinMode(alertPin, INPUT);

  // Wandlung starten
  writeRegister(ADS1X15_REG_POINTER_CONFIG, config);

  // Auslesen des Konvertierungsregisters vorbereiten
  _twi.beginTransmission(_i2cAddress);
  _twi.write(ADS1X15_REG_POINTER_CONVERT);
  _twi.endTransmission(true);

  attachInterrupt(digitalPinToInterrupt(alertPin), ISR, FALLING); // ISR registrieren

  _isContinuousMode = true;
}


void UrsAds1x15::stopContinuousMode() {
  if (!_isContinuousMode)
    return;

  detachInterrupt(_alertPin);
  // Config Register zusammenstellen
  uint16_t config = ADS1X15_REG_CONFIG_CQUE_NONE |    // Disable the comparator
                    ADS1X15_REG_CONFIG_CLAT_NONLAT |  // Non-latching
                    ADS1X15_REG_CONFIG_CPOL_ACTVLOW | // Alert/Rdy active low
                    ADS1X15_REG_CONFIG_CMODE_TRAD |   // Traditional comparator
                    ADS1X15_REG_CONFIG_MODE_SINGLE;

  writeRegister(ADS1X15_REG_POINTER_CONFIG, config);
  writeRegister(ADS1X15_REG_POINTER_HITHRESH, 0x7FFF); // Standardwerte lt. Datenblatt
  writeRegister(ADS1X15_REG_POINTER_LOWTHRESH, 0x8000);
  _isContinuousMode = false;
}


int16_t ICACHE_RAM_ATTR UrsAds1x15::readContinuousModeRaw() const {
  if (!_isContinuousMode) // kontinuierlicher Modus aktiv?
    return 0;

#ifdef __AVR__
 // Beim AVR erfolgt TWI per Interrupt. 
 // Wenn readContinuousModeRaw in einer ISR aufgerufen wird, sind die Interrupts abgeschaltet.
  interrupts();
#endif // __AVR__

  // Konvertierungsregister abrufen
  // Pointer-Register wurde bereits in startContinuousMode gesetzt.
  _twi.requestFrom(_i2cAddress, (uint8_t)2, (uint8_t)true); 
  noInterrupts();
  uint16_t res = ((_twi.read() << 8) | _twi.read());
  return (int16_t)res;
}


// ======================================================================================================
// ===  UrsAds1015Class
// ======================================================================================================
#if defined(ESP8266)
UrsAds1015::UrsAds1015(UrsTwi& twi, Ads1x15Address i2cAddress) : UrsAds1x15(twi, i2cAddress) {
#elif defined(__AVR__) 
UrsAds1015::UrsAds1015(Ads1x15Address i2cAddress) : UrsAds1x15(i2cAddress) {
#endif 
  setSps(Ads1015Sps::Sps1600);
}
 
void UrsAds1015::setSps(Ads1015Sps Sps) {
  _Sps = Sps;
  _SpsCode = (uint16_t)Sps;
  switch (Sps) {
  case Ads1015Sps::Sps128:
    _SamplesPerSecond = 128;
    _conversionDelay = 10;
    break;
  case Ads1015Sps::Sps250:
    _SamplesPerSecond = 250;
    _conversionDelay = 5;
    break;
  case Ads1015Sps::Sps490:
    _SamplesPerSecond = 490;
    _conversionDelay = 3;
    break;
  case Ads1015Sps::Sps920:
    _SamplesPerSecond = 920;
    _conversionDelay = 1;
    break;
  case Ads1015Sps::Sps1600:
    _SamplesPerSecond = 1600;
    _conversionDelay = 1;
    break;
  case Ads1015Sps::Sps2400:
    _SamplesPerSecond = 2400;
    _conversionDelay = 1;
    break;
  case Ads1015Sps::Sps3300:
    _SamplesPerSecond = 3300;
    _conversionDelay = 1;
    break;
  default: // Fehler
    _SamplesPerSecond = 0;
    _conversionDelay = 1;
    break;
  }
}


// ======================================================================================================
// ===  UrsAds1115Class
// ======================================================================================================
#if defined(ESP8266)
UrsAds1115::UrsAds1115(UrsTwi& twi, Ads1x15Address i2cAddress) : UrsAds1x15(twi, i2cAddress) {
#elif defined(__AVR__) 
UrsAds1115::UrsAds1115(Ads1x15Address i2cAddress) : UrsAds1x15(i2cAddress) {
#endif 
  setSps(Ads1115Sps::Sps128);
}

void UrsAds1115::setSps(Ads1115Sps Sps) {
  _Sps = Sps;
  _SpsCode = (uint16_t)Sps;
  switch (Sps) {
  case Ads1115Sps::Sps8:
    _SamplesPerSecond = 8;
    _conversionDelay = 150;
    break;
  case Ads1115Sps::Sps16:
    _SamplesPerSecond = 16;
    _conversionDelay = 100;
    break;
  case Ads1115Sps::Sps32:
    _SamplesPerSecond = 32;
    _conversionDelay = 40;
    break;
  case Ads1115Sps::Sps64:
    _SamplesPerSecond = 64;
    _conversionDelay = 20;
    break;
  case Ads1115Sps::Sps128:
    _SamplesPerSecond = 128;
    _conversionDelay = 8;
    break;
  case Ads1115Sps::Sps250:
    _SamplesPerSecond = 250;
    _conversionDelay = 5;
    break;
  case Ads1115Sps::Sps475:
    _SamplesPerSecond = 475;
    _conversionDelay = 3;
    break;
  case Ads1115Sps::Sps860:
    _SamplesPerSecond = 820;
    _conversionDelay = 2;
    break;
  default: // Fehler
    _SamplesPerSecond = 0;
    _conversionDelay = 1;
    break;
  }
}


Ads1x15Gain & operator++(Ads1x15Gain& p_oRight) { // ++A
  switch (p_oRight) {
  case Ads1x15Gain::Gain_2_3:
    p_oRight = Ads1x15Gain::Gain_1;
    break;
  case Ads1x15Gain::Gain_1:
    p_oRight = Ads1x15Gain::Gain_2;
    break;
  case Ads1x15Gain::Gain_2:
    p_oRight = Ads1x15Gain::Gain_4;
    break;
  case Ads1x15Gain::Gain_4:
    p_oRight = Ads1x15Gain::Gain_8;
    break;
  case Ads1x15Gain::Gain_8:
    p_oRight = Ads1x15Gain::Gain_16;
    break;
  case Ads1x15Gain::Gain_16:
    p_oRight = Ads1x15Gain::Gain_16;
    break;
  }
  return p_oRight;
}

Ads1x15Gain operator++(Ads1x15Gain & p_oRight, int) { // A++
  Ads1x15Gain oCopy = p_oRight;
  ++p_oRight;
  return oCopy;
}

Ads1x15Gain & operator--(Ads1x15Gain& p_oRight) { // --A
  switch (p_oRight) {
  case Ads1x15Gain::Gain_2_3:
    p_oRight = Ads1x15Gain::Gain_2_3;
    break;
  case Ads1x15Gain::Gain_1:
    p_oRight = Ads1x15Gain::Gain_2_3;
    break;
  case Ads1x15Gain::Gain_2:
    p_oRight = Ads1x15Gain::Gain_1;
    break;
  case Ads1x15Gain::Gain_4:
    p_oRight = Ads1x15Gain::Gain_2;
    break;
  case Ads1x15Gain::Gain_8:
    p_oRight = Ads1x15Gain::Gain_4;
    break;
  case Ads1x15Gain::Gain_16:
    p_oRight = Ads1x15Gain::Gain_8;
    break;
  }
  return p_oRight;
}

String UrsAds1x15::getGainString() {
  switch (_gain) {
  case Ads1x15Gain::Gain_2_3: return "2/3";
  case Ads1x15Gain::Gain_1: return "1";
  case Ads1x15Gain::Gain_2: return "2";
  case Ads1x15Gain::Gain_4: return "4";
  case Ads1x15Gain::Gain_8: return "8";
  case Ads1x15Gain::Gain_16: return "16";
  default:
    return "?";
  }
}

Ads1x15Gain operator--(Ads1x15Gain & p_oRight, int) { // A--
  Ads1x15Gain oCopy = p_oRight;
  --p_oRight;
  return oCopy;
}